#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2019, Lluís Vilanova"
__license__ = "GPL version 3 or later"


from contextlib import contextmanager
import joblib


@contextmanager
def step(message, logger=print):
    """Show simple progress messages around a piece of code.

    Parameters
    ----------
    message : str
        Message to print.
    logger : function, optinal
        Logging function. Defaults to `print`.

    Examples
    --------
    >>> with step("Doing something")
            print("some text")
    Doing something...
    some text
    Doing something... done

    """
    logger(message, "...")
    yield
    logger(message, "... done")


class threaded(object):
    """Context manager to run functions in parallel using threads.

    Examples
    --------
    Run two processes in parallel and wait until both are finished:

    >>> with step("Running in parallel"), threaded() as t:
            @t.start
            def f1():
                shell = LocalShell()
                shell.run(["sleep", "2"])
                print("f1")

            @t.start
            def f2():
                shell = LocalShell()
                shell.run(["sleep", "1"])
                print("f2")
    Running in parallel...
    f2
    f1
    Running in parallel... done

    """
    def __init__(self, n_jobs=None):
        if n_jobs is None:
            n_jobs = -1
        self._n_jobs = n_jobs
        self._jobs = []
        self.result = None

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pool = joblib.Parallel(backend="threading", n_jobs=self._n_jobs)
        self.result = pool(joblib.delayed(job, check_pickle=False)(*args, **kwargs)
        		   for job, args, kwargs in self._jobs)

    def start(self, target):
        """Decorator to start a function on a separate thread."""
        self._jobs.append((target, [], {}))

    def start_args(self, *args, **kwargs):
        """Callable decorator to start a function on a separate thread.

        Examples
        --------
        >>> with threaded() as t:
                @t.start_args(1, b=2)
                def f(a, b):
                    print(a, b)
        1, 2

        """
        def wrapper(target):
            self._jobs.append((target, args, kwargs))
        return wrapper


__all__ = [
    "step", "threaded",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""SciExp²-ExpRun provides a framework for easing the workflow of executing
experiments that require orchestrating multiple processes in local and/or remote
machines.

"""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2019, Lluís Vilanova"
__license__ = "GPL version 3 or later"


__version_info__ = (0, 1, 5)
__version__ = ".".join([str(i) for i in __version_info__])

__all__ = [
    "cpu", "files", "kernel", "process", "spur", "util", "wait",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Extensions to external `spur` package."""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2019, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import atexit
import collections
import logging
import os
import signal
import sys
import threading
import time
import traceback

import six
import spur
import spur.ssh


_LOGGER = logging.getLogger(__name__)


def is_local_shell(shell):
    """Whether the given shell is a `spur.LocalShell` or derivative."""
    return isinstance(shell, spur.LocalShell)


def is_ssh_shell(shell):
    """Whether the given shell is a `spur.SshShell` or derivative."""
    return isinstance(shell, spur.SshShell)


# Patch spur to update the _is_killed attribute when sending signals
def _patch_send_signal(func):
    def send_signal_wrapper(self, signum):
        if signum in [signal.SIGINT, signal.SIGQUIT, signal.SIGKILL]:
            # pylint: disable=protected-access
            self._is_killed = True
        return func(self, signum)
    return send_signal_wrapper
spur.ssh.SshProcess.send_signal = _patch_send_signal(spur.ssh.SshProcess.send_signal)


# Monitor background processes for failures, so we can error out early

_EXITING = False
_LOCK = threading.RLock()


def _kill_all():
    # pylint: disable=global-statement
    global _EXITING
    _EXITING = True
    # pylint: disable=protected-access
    LocalShell._atexit_cb()
    # pylint: disable=protected-access
    SshShell._atexit_cb()


atexit.register(_kill_all)


# pylint: disable=invalid-name,too-few-public-methods
class with_pid:
    """Decorator to define a kill argument that takes the process' pid.

    To be used as an element in the `kill` argument to a shell's `run` or
    `spawn` method::

        shell.run(["sudo", "program"], kill=["sudo", "kill", with_pid()])

    Can be used in three ways, depending on the type of `func`:
    - `None`: replace with the stringified process pid.
    - `str`: format with the process' pid with the ``pid`` key.
    - otherwise: call with the pid as an argument.

    """
    def __init__(self, func=None):
        self._func = func
    def __call__(self, pid):
        if self._func is None:
            return str(pid)
        if isinstance(self._func, six.string_types):
            return self._func.format(pid=pid)
        return self._func(pid)

def _kill_pid(shell, pid, kill_args):
    if kill_args:
        args = []
        for arg in kill_args:
            if isinstance(arg, with_pid):
                args.append(arg(pid))
            else:
                args.append(arg)
        shell.run(args)
    else:
        shell.run(["pkill", "-P", str(pid)])


def _print_traceback(cmd_msg, stack_info=None):
    if stack_info is None:
        stack_info = traceback.extract_stack()
    stack_idx = 0 if stack_info[0][2] == "<module>" else 6
    print("Traceback (most recent call last):")
    msg = traceback.format_list(stack_info[stack_idx:-1])
    print("".join(msg), end="")
    exc_type, exc_value, tb = sys.exc_info()
    info = traceback.extract_tb(tb)
    msg = traceback.format_list(info)
    print("".join(msg), end="")
    print("%s.%s: %s" % (exc_type.__module__, exc_type.__name__, exc_value))
    print("command:", cmd_msg)


def _watchdog_thread(shell, obj, cmd_msg, exit_on_error):
    stack_info = traceback.extract_stack()
    def watchdog():
        while obj.is_running():
            time.sleep(1)
        try:
            obj.wait_for_result()
        # pylint: disable=bare-except
        except:
            # pylint: disable=protected-access
            shell._child_remove(obj)
            _LOGGER.info("- %s", cmd_msg)
            # pylint: disable=protected-access
            if not obj._is_killed and not _EXITING:
                _print_traceback(cmd_msg, stack_info)
                if exit_on_error:
                    _kill_all()
                    os._exit(1)
        else:
            shell._child_remove(obj)    # pylint: disable=protected-access
            _LOGGER.info("- %s", cmd_msg)
    thread = threading.Thread(target=watchdog)
    thread.daemon = True
    thread.start()


# pylint: disable=function-redefined
class LocalShell(spur.LocalShell):
    """An extended version of `spur.LocalShell`.

    It will properly kill all created processes when exiting.

    The `run` and `spawn` methods have two new arguments:

    - ``exit_on_error``: bool, optional
          Whether to exit the program when the process fails to execute.

    - ``kill``: list of str, optional
          Command to execute when killing the process. Useful when process is
          run with sudo.

    """

    __CHILDREN = collections.OrderedDict()

    def _child_add(self, obj, kill):
        with _LOCK:
            LocalShell.__CHILDREN[obj] = (self, kill)

    @classmethod
    def _child_remove(cls, obj):
        with _LOCK:
            del cls.__CHILDREN[obj]

    # pylint: disable=arguments-differ
    def run(self, *args, **kwargs):
        process = self.spawn(*args, **kwargs, exit_on_error=False)
        result = process.wait_for_result()
        return result

    def spawn(self, *args, **kwargs):
        exit_on_error = kwargs.pop("exit_on_error", True)
        kill = kwargs.pop("kill", None)
        cmd = args[0]
        cmd_msg = " ".join(cmd)
        _LOGGER.info("+ %s", cmd_msg)
        try:
            obj = spur.LocalShell.spawn(self, *args, **kwargs, store_pid=True)
        # pylint: disable=bare-except
        except:
            if exit_on_error:
                stack_info = traceback.extract_stack()
                _print_traceback(cmd_msg, stack_info)
                _kill_all()
                # pylint: disable=protected-access
                os._exit(1)
            else:
                raise
        else:
            obj._is_killed = False      # pylint: disable=protected-access
            self._child_add(obj, kill)
            _watchdog_thread(self, obj, cmd_msg, exit_on_error)
            return obj

    @classmethod
    def _atexit_cb(cls):
        while True:
            with _LOCK:
                if not cls.__CHILDREN:
                    return
                child, (shell, kill) = next(iter(cls.__CHILDREN.items()))
                # pylint: disable=protected-access
                child._is_killed = True
                if child.is_running():
                    try:
                        _kill_pid(shell, child.pid, kill)
                        child.wait_for_result()
                    # pylint: disable=bare-except
                    except:
                        pass
                else:
                    try:
                        cls._child_remove(child)
                    # pylint: disable=bare-except
                    except:
                        pass


class SshShell(spur.SshShell):
    """An extended version of `spur.SshShell`.

    It will properly kill all created processes when exiting.

    The shell object has two new members:

    - ``hostname``: str
          Target host name.

    - ``username``: str
          Target user name.

    The `run` and `spawn` methods have two new arguments:

    - ``exit_on_error``: bool, optional
          Whether to exit the program when the process fails to execute.

    - ``kill``: list of str, optional
          Command to execute when killing the process. Useful when process is
          run with sudo.

    """

    __CHILDREN = collections.OrderedDict()

    def _child_add(self, obj, kill):
        with _LOCK:
            SshShell.__CHILDREN[obj] = (self, kill)

    @classmethod
    def _child_remove(cls, obj):
        with _LOCK:
            del cls.__CHILDREN[obj]

    def __init__(self, *args, **kwargs):
        spur.SshShell.__init__(self, *args, **kwargs)
        self.hostname = self._hostname
        self.username = self._username

    # pylint: disable=arguments-differ
    def run(self, *args, **kwargs):
        process = self.spawn(*args, **kwargs, exit_on_error=False)
        result = process.wait_for_result()
        return result

    # pylint: disable=arguments-differ
    def spawn(self, *args, **kwargs):
        exit_on_error = kwargs.pop("exit_on_error", True)
        kill = kwargs.pop("kill", None)
        cmd = args[0]
        cmd_msg = "ssh -p %d %s@%s %s" % (self._port, self.username, self.hostname, " ".join(cmd))
        _LOGGER.info("+ %s", cmd_msg)
        try:
            obj = spur.SshShell.spawn(self, *args, **kwargs, store_pid=True)
        # pylint: disable=bare-except
        except:
            if exit_on_error:
                stack_info = traceback.extract_stack()
                _print_traceback(cmd_msg, stack_info)
                _kill_all()
                # pylint: disable=protected-access
                os._exit(1)
            else:
                raise
        else:
            obj._is_killed = False      # pylint: disable=protected-access
            # pylint: disable=protected-access
            self._child_add(obj, kill)
            _watchdog_thread(self, obj, cmd_msg, exit_on_error)
            return obj

    @classmethod
    def _atexit_cb(cls):
        while True:
            with _LOCK:
                if not cls.__CHILDREN:
                    return
                child, (shell, kill) = next(iter(cls.__CHILDREN.items()))
                # pylint: disable=protected-access
                child._is_killed = True
                if child.is_running():
                    try:
                        _kill_pid(shell, child.pid, kill)
                        child.wait_for_result()
                    # pylint: disable=bare-except
                    except:
                        pass
                else:
                    try:
                        cls._child_remove(child)
                    # pylint: disable=bare-except
                    except:
                        pass


def get_shell(server, user=None, password=None, port=22):
    """Get a new shell.

    If `server` is a spur shell, return that instead.

    Parameters
    ----------
    server : str or object
    user : str, optional
    password : str, optional
    port : int, optional

    """
    if is_ssh_shell(server) or is_local_shell(server):
        return server
    return SshShell(hostname=server,
                    username=user,
                    password=password,
                    port=port,
                    missing_host_key=spur.ssh.MissingHostKey.accept)

__all__ = [
    "is_local_shell", "is_ssh_shell", "get_shell",
    "with_pid",
]

__all__ += spur.__all__



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2019, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import collections
import re
from . import kernel


def set_freq(shell, path="cpupower", ld_library_path="", freq="max"):
    """Set frequency scaling.

    Parameters
    ----------
    shell
        Target shell.
    path : str, optional
        Path to cpupower tool. Default is use the cpupower tool in the PATH.
    ld_library_path : str, optional
        Library path to run cpupower tool. Default is use the system's library
        path.
    freq : str, optional
        Frequency to set in GHz. Default is use maximum frequency.

    Notes
    -----
    In some systems it might be necessary to boot the Linux kernel with command
    line option "intel_pstate=disable" in order to support user frequency
    settings.

    """
    if freq == "max":
        max_freq = shell.run([
            "sh", "-c",
            f"sudo LD_LIBRARY_PATH={ld_library_path} {path} frequency-info | grep 'hardware limits' | sed -e 's/.* - \\(.*\\) GHz/\\1/'"],
                             encoding="ascii")
        freq = max_freq.output[:-1]

    shell.run(["sudo",
               f"LD_LIBRARY_PATH={ld_library_path}", path,
               "-c", "all", "frequency-set", "--governor", "userspace"])
    shell.run(["sudo",
               f"LD_LIBRARY_PATH={ld_library_path}", path,
               "-c", "all", "frequency-set", "--freq", f"{freq}GHz"])


def _get_mask(cpu_list):
    mask = 0
    for cpu in cpu_list:
        mask += 1 << cpu
    return mask

def set_irqs(shell, *irqs, **kwargs):
    """Make irqbalance ignore the given IRQs, and instead set their SMP affinity.

    Parameters
    ----------
    shell
        Target system.
    irqs
        IRQ descriptors.
    ignore_errors : bool, optional
        Ignore errors when manually setting an IRQ's SMP affinity. Implies that
        irqbalance will manage that IRQ. Default is False.
    irqbalance_banned_cpus : list of int, optional
        CPUs that irqbalance should not use for balancing.
    irqbalance_args : list of str, optional
        Additional arguments to irqbalance.

    Each descriptor in `irqs` is a three-element tuple:
    * Type: either ``irq`` for the first column in /proc/interrupts, or
            ``descr`` for the interrupt description after the per-CPU counts.
    * Regex: a regular expression to apply to the fields above, or `True` to
             apply to all values (a shorthand to the regex ".*"), or an `int` (a
             shorthand to the regex "^int_value$").
    * SMP affinity: list of cpu numbers to set as the IRQ's affinity; if `True`
                    is used instead, let irqbalance manage this IRQ.

    All matching descriptors are applied in order for each IRQ. If no descriptor
    matches, or the last matching descriptor has `True` as its affinity value,
    the IRQ will be managed by irqbalance as before.

    Returns
    -------
    The new irqbalance process.

    """
    ignore_errors = kwargs.pop("ignore_errors", False)
    irqbalance_args = kwargs.pop("irqbalance_args", [])
    irqbalance_banned_cpus = kwargs.pop("irqbalance_banned_cpus", [])
    irqbalance_banned_cpus_mask = _get_mask(irqbalance_banned_cpus)
    if len(kwargs) > 0:
        raise Exception("unknown argument: %s" % list(kwargs.keys())[0])

    irqs_parsed = []
    for arg_irq in irqs:
        if len(arg_irq) != 3:
            raise ValueError("wrong IRQ descriptor: %s" % repr(arg_irq))

        irq_type, irq_re, irq_cpus = arg_irq

        if isinstance(irq_re, int):
            irq_re = "^%d$" % irq_re
        if not isinstance(irq_re, bool) and not isinstance(irq_re, six.string_types):
            raise TypeError("wrong IRQ descriptor regex: %s" % str(irq_re))
        if not isinstance(irq_re, bool):
            irq_re = re.compile(irq_re)

        if (not isinstance(irq_cpus, bool) and (isinstance(irq_cpus, six.string_types) or
                                              not isinstance(irq_cpus, collections.Iterable))):
            raise TypeError("wrong IRQ descriptor CPU list: %s" % str(irq_cpus))

        if irq_type not in ["irq", "descr"]:
            raise ValueError("wrong IRQ descriptor type: %s" % str(irq_type))

        irqs_parsed.append((irq_type, irq_re, irq_cpus))

    irq_manual = []
    irqbalance_banned = set()

    cre = re.compile(r"(?P<irq>[^:]+):(?:\s+[0-9]+)+\s+(?P<descr>.*)")
    with shell.open("/proc/interrupts") as f:
        for line in f.read().split("\n"):
            match = cre.match(line)
            if match is None:
                continue

            irq = match.groupdict()["irq"].strip()
            descr = match.groupdict()["descr"].strip()

            cpus = True

            for irq_type, irq_cre, irq_cpus in irqs_parsed:
                if irq_type == "irq":
                    if irq_cre == True or irq_cre.match(irq):
                        cpus = irq_cpus
                elif irq_type == "descr":
                    if irq_cre == True or irq_cre.match(descr):
                        cpus = irq_cpus
                else:
                    assert False, irq_type

            if cpus != True:
                irq_manual.append((irq, cpus))
                irqbalance_banned.add(irq)

    for irq, cpus in irq_manual:
        mask = _get_mask(cpus)
        try:
            shell.run(["sudo", "sh", "-c",
                       "echo %x > /proc/irq/%s/smp_affinity" % (irq, mask)])
        except:
            if ignore_errors:
                irqbalance_banned.remove(irq)
            else:
                raise

    shell.run(["sudo", "service", "irqbalance", "stop"])
    proc = shell.spawn(["sudo", "IRQBALANCE_BANNED_CPUS=%x" % irqbalance_banned_cpus_mask,
                        "irqbalance"] + irqbalance_args +
                       ["--banirq=%s" % banned
                        for banned in irqbalance_banned],
                       encoding="ascii")
    return proc


def get_cpus(shell, node=None, package=None, core=None, pu=None, cgroup=None):
    """Get a set of all physical CPU indexes in the system.

    It uses the hwloc-calc program to report available CPUs.

    Parameters
    ----------
    shell
        Target shell.
    node : int or str, optional
        NUMA nodes to check. Defaults to all.
    package : int or str, optional
        Core packages to check on selected NUMA nodes. Defaults to all.
    core : int or str, optional
        Cores to check on selected core packages. Defaults to all.
    pu : int or str, optional
        PUs to check on selected cores. Defaults to all.
    cgroup : str, optional
        Cgroup path.

    Returns
    -------
    set of int
        Physical CPU indexes (as used by Linux).

    Notes
    -----
    The combination of all the arguments is a flexible way to get all the
    information in the system. Each of these arguments can have any of the forms
    described under "Hwloc Indexes" in manpage hwloc(7). A few examples.

    Second thread of each core:
    >>> get_cpus(shell, pu=1)

    First thread of each core in first NUMA node:
    >>> get_cpus(node=0, pu=0)

    Hardware threads in first core of the entire system:
    >>> get_cpus(node=0, package=0, core=0)

    """
    cmd = ["hwloc-calc", "--intersect", "PU",
           "--li", "--po", ""]

    def add_level(name, value):
        if value is not None:
            cmd[-1] += ".%s:%s" % (name, str(value))
    add_level("node", node)
    add_level("package", package)
    add_level("core", core)
    add_level("pu", pu)
    if cmd[-1] == "":
        cmd[-1] = "all"
    else:
        cmd[-1] = cmd[-1][1:]

    if cgroup is not None:
        cmd = ["sudo", "cgexec", "-g", cgroup] + cmd

    res = shell.run(cmd, encoding="ascii")
    line = res.output.split("\n")[0]
    if line == "":
        raise ValueError("hwloc-calc: %r" % res.stderr_output)
    return [int(i) for i in res.output.split(",")]


__all__ = [
    "set_freq", "set_irqs", "get_cpus",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2019, Lluís Vilanova"
__license__ = "GPL version 3 or later"


from . import spur


def install(shell, package):
    """Install given `package` using `shell`."""
    if spur.is_ssh_shell(shell):
        hostname = shell.hostname
    else:
        hostname = "localhost"
    shell.run([
        "bash", "-c",
        "dpkg -s %s >/dev/null 2>&1 || sudo apt-get install -y %s" % (package,
                                                                      package),
    ])


def install_deps(shell):
    """Install all needed system packages.

    Must be called on a local shell before using other functions that require a
    shell, and before using other functions through the same shell.

    Parameters
    ----------
    shell
        Target system.

    """
    install(shell, "cgroup-tools")
    install(shell, "hwloc")
    install(shell, "rsync")
    install(shell, "netcat-traditional")
    install(shell, "psmisc")
    install(shell, "util-linux")


def rsync(src_shell, src_path, dst_shell, dst_path, run_shell=None, args=[]):
    """Synchronize two directories using rsync.

    Parameters
    ----------
    src_shell
        Source shell.
    src_path
        Source directory.
    dst_shell
        Destination shell.
    dst_path
        Destination directory.
    run_shell : optional
        Shell where to run rsync. Default is local machine.
    args : list of str, optional
        Additional arguments to rsync. Default is none.

    """
    if (not spur.is_local_shell(src_shell) and not spur.is_local_shell(dst_shell) and
        run_shell is not src_shell and run_shell is not dst_shell):
        raise Exception("rsync cannot work with two remote shells")

    if run_shell is None:
        run_shell = spur.LocalShell()

    ssh_port = 22
    cmd_pass = []
    if spur.is_local_shell(src_shell) or run_shell is src_shell:
        cmd_src = [src_path]
    else:
        ssh_port = src_shell._port
        if src_shell._password is not None:
            cmd_pass = ["sshpass", "-p", src_shell._password]
        cmd_src = ["%s@%s:%s" % (src_shell.username, src_shell.hostname, src_path)]
    if spur.is_local_shell(dst_shell) or run_shell is dst_shell:
        cmd_dst = [dst_path]
    else:
        ssh_port = dst_shell._port
        if dst_shell._password is not None:
            cmd_pass = ["sshpass", "-p", dst_shell._password]
        cmd_dst = ["%s@%s:%s" % (dst_shell.username, dst_shell.hostname, dst_path)]

    cmd = []
    cmd += cmd_pass
    cmd += ["rsync", "-az"]
    cmd += ["-e", "ssh -p %d -o StrictHostKeyChecking=no" % ssh_port]
    cmd += cmd_src
    cmd += cmd_dst
    cmd += args
    run_shell.run(cmd)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2019, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import logging
from . import wait

logger = logging.getLogger(__name__)


def check_version(shell, version, fail=True):
    """Check that a specific linux kernel version is installed.

    Parameters
    ----------
    shell
        Target shell.
    version : str
        Target kernel version.
    fail : bool, optional
        Whether to raise an exception when a different version is
        installed. Default is True.

    Returns
    -------
    bool
        Whether the target kernel version is installed.

    """
    res = shell.run(["uname", "-r"])
    current = res.output.split("\n")[0]
    if current == version:
        return True
    else:
        if fail:
            raise Exception("Invalid kernel version: target=%s current=%s" % (version, current))
        return False


def install_version(shell, version, package_base):
    """Install and reboot into a given linux kernel version if it is not the current.

    Parameters
    ----------
    shell
        Target shell.
    version : str
        Target kernel version.
    package_base : str
        Base directory in target shell where kernel packages can be installed
        from.

    """
    if check_version(shell, version, fail=False):
        return

    for name in ["linux-image-%(v)s_%(v)s-*.deb",
                 "linux-headers-%(v)s_%(v)s-*.deb",
                 "linux-libc-dev_%(v)s-*.deb"]:
        name = os.path.join(package_base, name % {"v": version})
        res = shell.run(["sh", "-c", "ls %s" % name])
        files = res.output.split("\n")
        for path in files:
            if path == "":
                continue
            logger.warn("Installing %s..." % path)
            shell.run(["sudo", "dpkg", "-i", path])

    res = shell.run(["grep", "-E", "menuentry .* %s" % version, "/boot/grub/grub.cfg"])
    grub_ids = res.output.split("\n")
    pattern = r" '([a-z0-9.-]+-%s-[a-z0-9.-]+)' {" % re.escape(version)
    grub_id = re.search(pattern, grub_ids[0]).group(1)

    with step("Updating GRUB %s..." % path, logger=logger.warn):
        shell.run(["sudo", "sed", "-i", "-e",
                   "s/^GRUB_DEFAULT=/GRUB_DEFAULT=\"saved\"/",
                   "/etc/default/grub"])
        shell.run(["sudo", "update-grub"])
        shell.run(["sudo", "grub-set-default", grub_id])

    with step("Rebooting into new kernel...", logger=logger.warn):
        shell.run(["sudo", "reboot"], allow_error=True)
        wait.ssh(shell)

    check_version(shell, version)


def check_cmdline(shell, arg):
    """Check the linux kernel was booted with the given commandline.

    Parameters
    ----------
    shell
        Target shell.
    arg : str
        Command line argument to check.

    """
    shell.run(["grep", arg, "/proc/cmdline"])


def check_module_param(shell, module, param, value, fail=True):
    """Check that a linux kernel module was loaded with the given parameter value.

    Parameters
    ----------
    shell
        Target shell.
    module : str
        Module name.
    param : str
        Module name.
    value
        Module value (will be coverted to str).
    fail : bool, optional
        Raise an exception if the value is not equal. Default is True.

    Returns
    -------
    bool
        Whether the given kernel module was loaded with the given parameter
        value.

    """
    with shell.open("/sys/module/%s/parameters/%s" % (module, param), "r") as f:
        f_val = f.read().split("\n")[0]
        if f_val != value:
            if fail:
                raise Exception("invalid kernel parameter value: target=%s current=%s" % (value, f_val))
            return False
        else:
            return True


__all__ = [
    "check_version", "install_version", "check_cmdline",
    "check_module_param",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Process management functions."""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2019, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import collections
import os
import re
import six


# pylint: disable=redefined-builtin
def get_tids(shell, pid, filter=None):
    """Get ids of all threads in a given process.

    Parameters
    ----------
    shell
        Target shell.
    pid : int
        Target process pid.
    filter : str or cre, optional
        Return pids that match given filter in process name. Default is all
        pids.

    Returns
    -------
    list of int
        List of the selected process pids.

    Notes
    -----
    When using a string for `filter` it will simply check it is part of the
    process name.

    """
    pids = shell.run(["ps", "H", "-o", "tid comm", str(pid)])
    lines = pids.output.split("\n")
    res = []
    for line in lines[1:]:
        line = line.strip()
        if line == "":
            continue
        pid, name = line.split(" ", 1)
        if filter:
            if isinstance(filter, six.string_types):
                if filter not in name:
                    continue
            else:
                if not filter.match(name):
                    continue
        res.append(int(pid))
    return res


def pin(shell, pid, cpus, **shell_kwargs):
    """Pin pid to given physical CPU list.

    Parameters
    ----------
    shell
        Target shell.
    pid : int
        Target pid or tid to pin.
    cpus : list of int
        Physical CPUs to pin the pid to.
    shell_kwargs : optional
        Arguments to `shell.run`

    """
    shell.run(["sudo", "taskset", "-p",
               "-c", ",".join(str(c) for c in cpus), str(pid)],
              **shell_kwargs)


def cgroup_create(shell, controller, path, **kwargs):
    """Create a cgroup for given subsystem.

    Parameters
    ----------
    shell
        Target shell.
    controller : str
        Cgroup controller to configure.
    path : str
        New cgroup path.
    kwargs : dict
        Controller parameters to set. Lists are comma-concatenated, all elements
        are transformed to str.

    """
    shell.run(["sudo", "cgcreate", "-g", controller+":"+path])
    for key, val in kwargs.items():
        if isinstance(val, six.string_types) or not isinstance(val, collections.Iterable):
            val = [val]
        val = ",".join(str(v) for v in val)
        shell.run(["sudo", "cgset", "-r", "%s.%s=%s" % (controller, key, val), path])


def cgroup_pids(shell, path=None):
    """Get pids in given cgroup path.

    Parameters
    ----------
    shell
        Target shell.
    path : str, optional
        Cgroup path to analyze (defaults to entire system).

    Returns
    -------
    list of int
        Pids in the given cgroup.

    """
    res = set()
    base = "/sys/fs/cgroup"
    if path is None:
        path = ""
    cre = re.compile(os.path.join(base, "[^/]*", path))
    proc = shell.run(["find", base, "-name", "tasks"], encoding="ascii")
    for filepath in proc.output.split("\n"):
        if cre.match(os.path.dirname(filepath)):
            # pylint: disable=invalid-name
            with shell.open(filepath, "r") as f:
                pids = (int(pid) for pid in f.read().split("\n")
                        if pid != "")
                res.update(set(pids))
    return list(res)


def cgroup_move(shell, controller, path, pids):
    """Move pids to a cgroup.

    Parameters
    ----------
    shell
        Target shell.
    controller : str
        Cgroup controller.
    path : str
        Cgroup path.
    pids : pid or list of pid
        Pids to move into the cgroup. All elements are transformed to str.

    Notes
    -----
    If you move the process that is serving this shell, you might have to
    reconnect the shell to continue using it.

    """
    if isinstance(pids, six.string_types) or not isinstance(pids, collections.Iterable):
        pids = [pids]
    pids_str = " ".join(str(p) for p in pids if str(p) != "")
    shell.run([
        "sh", "-c",
        f"for p in {pids_str}; do sudo cgclassify -g {controller}:{path} $p || true; done"])

__all__ = [
    "get_tids", "pin",
    "cgroup_create", "cgroup_pids", "cgroup_move",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Functions to wait on remote connections and process output."""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2019, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import re
import io
import time
from . import spur


def run(shell, *args, **kwargs):
    """Run command with a timeout.

    Parameters
    ----------
    shell
       Shell used to run given command.
    timeout : int, optional
       Timeout before erroring out (in seconds). Default is no timeout.
    rerun_error : bool, optional
       Rerun command every time it fails. Default is False.
    args, kwargs
       Paramaters to the shell's spawn method.

    Returns
    -------
    spur.ExecutionResult

    """
    timeout = kwargs.pop("timeout", 0)
    rerun_error = kwargs.pop("rerun_error", False)
    allow_error = kwargs.pop("allow_error", False)

    proc = None
    t_start = time.time()

    while True:
        t_now = time.time()
        if t_now - t_start > timeout > 0:
            raise Exception("Wait timed out" + repr((t_now - t_start, timeout)))
        if proc is None:
            proc = shell.spawn(*args, allow_error=True, **kwargs)
        if proc.is_running():
            time.sleep(2)
        else:
            res = proc.wait_for_result()
            if res.return_code == 0:
                return res
            if not allow_error:
                if rerun_error:
                    proc = None
                    time.sleep(2)
                else:
                    raise res.to_error()
            else:
                return res


def connection(shell, address, port, timeout=0):
    """Wait until we can connect to given address/port."""
    cmd = ["sh", "-c", "echo | nc %s %d" % (address, port)]
    run(shell, cmd, timeout=timeout, rerun_error=True)


def ssh(shell, timeout=0):
    """Wait until we can ssh through given shell."""
    if spur.is_local_shell(shell):
        return
    local = spur.LocalShell()
    cmd = [
        # pylint: disable=protected-access
        "sshpass", "-p", shell._password,
        "ssh",
        "-o", "ConnectTimeout=1",
        "-o", "StrictHostKeyChecking=no",
        # pylint: disable=protected-access
        "-p", str(shell._port), shell.username+"@"+shell.hostname,
        "true",
    ]
    run(local, cmd, timeout=timeout, rerun_error=True)


def print_stringio(obj, out=None):
    """Print contents of a StringIO object as they become available.

    Useful in combination with `stringio` to print an output while processing
    it.

    Parameters
    ----------
    obj : StringIO
        StringIO object to print.
    out : file or function, optional
        File or function to print object's contents. Defaults to stdout.

    Examples
    --------
    >>> stdout = StringIO.StringIO()
    >>> thread.start_new_thread(print_stringio, (stdout,))
    >>> shell.run(["sh", "-c", "sleep 1 ; echo start ; sleep 2; echo end ; sleep 1"],
    ...           stdout=stdout)
    start
    end

    See also
    --------
    stringio

    """
    if not isinstance(obj, io.StringIO):
        raise TypeError("expected a StringIO object")
    if callable(out):
        def flush():
            pass
    else:
        def flush():
            if out is not None:
                out.flush()
        # pylint: disable=invalid-name
        def f(message):
            print(message, end="", out=out)
        out = f
    seen = 0
    while True:
        time.sleep(0.5)
        contents = obj.getvalue()
        missing = contents[seen:]
        if missing:
            out(missing)
            flush()
            seen += len(missing)


def stringio(obj, pattern):
    """Wait until a StringIO's contents match the given regex.

    Useful to trigger operations when a process generates certain output.

    Examples
    --------
    Count time between the "start" and "end" lines printed by a process:

    >>> stdout = io.StringIO()
    >>> def timer(obj):
            stringio(obj, "^start$")
            t_start = time.time()
            stringio(obj, "^end$")
            t_end = time.time()
            print("time:", int(t_end - t_start))
    >>> thread.start_new_thread(timer, (stdout,))
    >>> shell.run(["sh", "-c", "sleep 1 ; echo start ; sleep 2; echo end ; sleep >>> 1"],
    ...           stdout=stdout)
    time: 2

    See also
    --------
    print_stringio

    """
    if not isinstance(obj, io.StringIO):
        raise TypeError("expected a StringIO object")
    cre = re.compile(pattern, re.MULTILINE)
    while True:
        time.sleep(0.5)
        contents = obj.getvalue()
        match = cre.findall(contents)
        if match:
            return


__all__ = [
    "run", "connection", "ssh", "print_stringio", "stringio",
]
